from random import Random
from typing import Optional

from rdkit import Chem, rdBase
from rdkit.Chem import AllChem

rdBase.DisableLog("rdApp.error")


def cut(mol, rng: Random):
    if not mol.HasSubstructMatch(Chem.MolFromSmarts("[*]-;!@[*]")):
        return None

    bis = rng.choice(mol.GetSubstructMatches(Chem.MolFromSmarts("[*]-;!@[*]")))  # single bond not in ring

    bs = [mol.GetBondBetweenAtoms(bis[0], bis[1]).GetIdx()]

    fragments_mol = Chem.FragmentOnBonds(mol, bs, addDummies=True, dummyLabels=[(1, 1)])

    try:
        return Chem.GetMolFrags(fragments_mol, asMols=True, sanitizeFrags=True)
    except ValueError:
        return None


def cut_ring(mol, rng: Random):
    for _ in range(10):
        if rng.random() < 0.5:
            if not mol.HasSubstructMatch(Chem.MolFromSmarts("[R]@[R]@[R]@[R]")):
                return None
            bis = rng.choice(mol.GetSubstructMatches(Chem.MolFromSmarts("[R]@[R]@[R]@[R]")))
            bis = (
                (bis[0], bis[1]),
                (bis[2], bis[3]),
            )
        else:
            if not mol.HasSubstructMatch(Chem.MolFromSmarts("[R]@[R;!D2]@[R]")):
                return None
            bis = rng.choice(mol.GetSubstructMatches(Chem.MolFromSmarts("[R]@[R;!D2]@[R]")))
            bis = (
                (bis[0], bis[1]),
                (bis[1], bis[2]),
            )

        bs = [mol.GetBondBetweenAtoms(x, y).GetIdx() for x, y in bis]

        fragments_mol = Chem.FragmentOnBonds(mol, bs, addDummies=True, dummyLabels=[(1, 1), (1, 1)])

        try:
            fragments = Chem.GetMolFrags(fragments_mol, asMols=True, sanitizeFrags=True)
            if len(fragments) == 2:
                return fragments
        except ValueError:
            return None

    return None


def ring_OK(mol):
    if not mol.HasSubstructMatch(Chem.MolFromSmarts("[R]")):
        return True

    ring_allene = mol.HasSubstructMatch(Chem.MolFromSmarts("[R]=[R]=[R]"))

    cycle_list = mol.GetRingInfo().AtomRings()
    max_cycle_length = max([len(j) for j in cycle_list])
    macro_cycle = max_cycle_length > 6

    double_bond_in_small_ring = mol.HasSubstructMatch(Chem.MolFromSmarts("[r3,r4]=[r3,r4]"))

    return not ring_allene and not macro_cycle and not double_bond_in_small_ring


def mol_ok(
    mol,
    rng: Random,
    min_num_atoms=1,
    mean_num_atoms=40.0,
    std_num_atoms=1e3,  # by default, no real limit on maximum mol size
):
    try:
        Chem.SanitizeMol(mol)
        target_size = rng.gauss(mean_num_atoms, std_num_atoms)
        if mol.GetNumAtoms() > min_num_atoms and mol.GetNumAtoms() < target_size:
            return True
        else:
            return False
    except ValueError:
        return False


def crossover_ring(parent_A, parent_B, rng: Random, **mol_ok_kwargs):
    ring_smarts = Chem.MolFromSmarts("[R]")
    if not parent_A.HasSubstructMatch(ring_smarts) and not parent_B.HasSubstructMatch(ring_smarts):
        return None

    rxn_smarts1 = [
        "[*:1]~[1*].[1*]~[*:2]>>[*:1]-[*:2]",
        "[*:1]~[1*].[1*]~[*:2]>>[*:1]=[*:2]",
    ]
    rxn_smarts2 = [
        "([*:1]~[1*].[1*]~[*:2])>>[*:1]-[*:2]",
        "([*:1]~[1*].[1*]~[*:2])>>[*:1]=[*:2]",
    ]

    for i in range(10):
        fragments_A = cut_ring(parent_A, rng)
        fragments_B = cut_ring(parent_B, rng)

        if fragments_A is None or fragments_B is None:
            return None

        new_mol_trial = []  # type: ignore  # wants list type
        for rs in rxn_smarts1:
            rxn1 = AllChem.ReactionFromSmarts(rs)
            new_mol_trial = []
            for fa in fragments_A:
                for fb in fragments_B:
                    new_mol_trial.append(rxn1.RunReactants((fa, fb))[0])

        new_mols = []
        for rs in rxn_smarts2:
            rxn2 = AllChem.ReactionFromSmarts(rs)
            for m in new_mol_trial:
                m = m[0]
                if mol_ok(m, rng, **mol_ok_kwargs):
                    new_mols += list(rxn2.RunReactants((m,)))

        new_mols2 = []
        for m in new_mols:
            m = m[0]
            if mol_ok(m, rng=rng, **mol_ok_kwargs) and ring_OK(m):
                new_mols2.append(m)

        if len(new_mols2) > 0:
            return rng.choice(new_mols2)

    return None


def crossover_non_ring(parent_A, parent_B, rng: Random, **mol_ok_kwargs):
    for i in range(10):
        fragments_A = cut(parent_A, rng)
        fragments_B = cut(parent_B, rng)
        if fragments_A is None or fragments_B is None:
            return None
        rxn = AllChem.ReactionFromSmarts("[*:1]-[1*].[1*]-[*:2]>>[*:1]-[*:2]")
        new_mol_trial = []
        for fa in fragments_A:
            for fb in fragments_B:
                new_mol_trial.append(rxn.RunReactants((fa, fb))[0])

        new_mols = []
        for mol in new_mol_trial:
            mol = mol[0]
            if mol_ok(mol, rng=rng, **mol_ok_kwargs):
                new_mols.append(mol)

        if len(new_mols) > 0:
            return rng.choice(new_mols)

    return None


def crossover(parent_A, parent_B, rng: Random, **mol_ok_kwargs) -> Optional[Chem.Mol]:
    parent_smiles = [Chem.MolToSmiles(parent_A), Chem.MolToSmiles(parent_B)]
    try:
        Chem.Kekulize(parent_A, clearAromaticFlags=True)
        Chem.Kekulize(parent_B, clearAromaticFlags=True)

    except ValueError:
        pass

    for _ in range(10):
        if rng.random() <= 0.5:
            new_mol = crossover_non_ring(parent_A, parent_B, rng=rng, **mol_ok_kwargs)
            if new_mol is not None:
                new_smiles = Chem.MolToSmiles(new_mol)
                if new_smiles is not None and new_smiles not in parent_smiles:
                    return new_mol
        else:
            new_mol = crossover_ring(parent_A, parent_B, rng=rng, **mol_ok_kwargs)
            if new_mol is not None:
                new_smiles = Chem.MolToSmiles(new_mol)
                if new_smiles is not None and new_smiles not in parent_smiles:
                    return new_mol

    return None



"""Maximize any number of toy oracle functions."""

from __future__ import annotations

import argparse
import logging
import time

import joblib
import numpy as np

try:
    from tdc import Oracle
except ImportError:
    print("This script requires the TDC package to be installed.")
    exit()

from mol_ga.mol_libraries import random_zinc
from mol_ga.preconfigured_gas import default_ga

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument("--oracle", type=str, default="qed")
    parser.add_argument("--offspring_size", type=int, default=1000)
    parser.add_argument("--max_generations", type=int, default=100)
    parser.add_argument("--start_population_size", type=int, default=10_000)
    args = parser.parse_args()
    logging.info(f"Arguments: {args}")
    logging.info(f"Running GA to maximize {args.oracle}...")

    oracle = Oracle(name=args.oracle)

    start_time = time.monotonic()
    with joblib.Parallel(n_jobs=-1) as parallel:
        output = default_ga(
            starting_population_smiles=random_zinc(args.start_population_size),
            scoring_function=oracle,
            max_generations=args.max_generations,
            offspring_size=args.offspring_size,
            parallel=parallel,
        )
    end_time = time.monotonic()
    top_100_scores = sorted([score for score, _ in output.population], reverse=True)[:100]
    print("Top few scores:")
    print(top_100_scores[:25])
    print(f"Time elapsed: {end_time - start_time:.2f} seconds")
    guacamol_top1_10_100_score = (top_100_scores[0] + np.mean(top_100_scores[:10]) + np.mean(top_100_scores[:100])) / 3
    print(
        f"Average of top 1, top 10, and top 100 scores (reported metric in Guacamol): {guacamol_top1_10_100_score:.2f}"
    )



from .general_ga import run_ga_maximization
from .preconfigured_gas import default_ga

__all__ = ["run_ga_maximization", "default_ga"]

try:
    from importlib.metadata import PackageNotFoundError, version

    try:
        __version__ = version("mol_ga")
    except PackageNotFoundError:
        # package is not installed
        pass
except ModuleNotFoundError:
    pass  # Python < 3.8



from __future__ import annotations

import heapq

from mol_ga.general_ga import GAResults, run_ga_maximization
from mol_ga.graph_ga.gen_candidates import graph_ga_blended_generation
from mol_ga.sample_population import uniform_quantile_sampling


def default_ga(
    starting_population_smiles: list[str],
    scoring_function,
    max_generations: int,
    offspring_size: int,
    offspring_gen_func=graph_ga_blended_generation,
    population_sampling_function=uniform_quantile_sampling,
    population_size=10_000,
    **kwargs,
) -> GAResults:
    """Genetic algorithm with default parameters for maximizing a scoring function."""
    return run_ga_maximization(
        starting_population_smiles=set(starting_population_smiles),
        scoring_func=scoring_function,
        offspring_gen_func=offspring_gen_func,
        sampling_func=population_sampling_function,
        selection_func=heapq.nlargest,
        max_generations=max_generations,
        population_size=population_size,
        offspring_size=offspring_size,
        **kwargs,
    )
